@@ -8,7 +8,19 @@ from slvcodec import entity, typs, package_generator, config, vhdl_parser
 logger = logging.getLogger(__name__)
 
 
-def make_generics_wrapper(enty, generics, wrapped_name, ports_to_remove=None, for_arch_header='', slv_interface=True):
+def make_generics_wrapper(enty, generics, wrapped_name, ports_to_remove=None, for_arch_header='',
+                          slv_interface=True):
+    """
+    Create a wrapper around an entity which sets the generic parameters.
+    Args:
+      `enty`: A resolved entity object parsed from the VHDL.
+      `generics`: A dictionary of generics to set.
+      `wrapped_name`: The name of the entity that will do the wrapping.
+      `ports_to_remove`: These ports will not be connected to the wrapper. 
+          Use for output ports that we don't want to expose in the wrapper.
+      `for_arch_header`: Text placed at the top of the architecture.  Used for encryption headers.
+      `slv_interface`: Convert all ports to std_logic_vector and std_logic in the wrapper's ports.
+    """
     if ports_to_remove is None:
         ports_to_remove = []
     if for_arch_header is None:
@@ -18,7 +30,7 @@ def make_generics_wrapper(enty, generics, wrapped_name, ports_to_remove=None, fo
         if isinstance(v, str) and (len(v) > 0) and (v[0] not in  ("'", '"')):
             generics[k] = '"' + v + '"'
     # Get the list of generic parameters for the testbench.
-    wrapped_generics = ',\n'.join(['{} => {}'.format(g.name, generics[g.name])
+    wrapped_generics = ',\n'.join(['{} => {}'.format(g.name, g.name)
                                    for g in enty.generics.values()])
     # Generate use clauses required by the testbench.
     use_clauses = '\n'.join([
@@ -47,10 +59,31 @@ def make_generics_wrapper(enty, generics, wrapped_name, ports_to_remove=None, fo
         wrapper_ports=list([e for e in enty.ports.values() if e.name not in ports_to_remove]),
         for_arch_header=for_arch_header,
         )
-    return wrapper
+    template_name = 'setgenerics_pkg.vhd'
+    template_fn = os.path.join(os.path.dirname(__file__), 'templates', template_name)
+    with open(template_fn, 'r') as f:
+        template = jinja2.Template(f.read())
+    constant_generics = [
+        (g.name, str(g.typ), generics[g.name])
+        for g in enty.generics.values()]
+    package = template.render(
+        use_clauses=use_clauses,
+        generics=constant_generics,
+        )
+    return wrapper, package
 
 
 def make_double_wrapper(enty, default_generics=None):
+    """
+    Create a two wrappers around an entity.
+    The first wrapper converts all the ports to std_logic_vector and std_logic.
+    The second wrapper converts them back to the original types.
+    Between these two wrappers is then a convenient place to do synthesis and the output wrapper
+    can be used for testing.
+    Args:
+      `enty`: A resolved entity object parsed from the VHDL.
+      `default_generics`: A dictionary of generics to set as default in the wrappers.
+    """
     if default_generics is None:
         default_generics = {}
     else:
@@ -78,7 +111,8 @@ def make_double_wrapper(enty, default_generics=None):
     use_clauses += '\n' + '\n'.join([
         'use {}.{}_slvcodec.{};'.format(u.library, u.design_unit, u.name_within)
         for u in enty.uses.values()
-        if u.library not in ('ieee', 'std') and '_slvcodec' not in u.design_unit])
+        if u.library not in ('ieee', 'std') and ('_slvcodec' not in u.design_unit) and
+        (u.design_unit != 'slvcodec')])
     # Read in the toslvcodec template and format it.
     template_and_wrapped_names = (
         ('fromslvcodec.vhd', enty.identifier),
@@ -101,15 +135,38 @@ def make_double_wrapper(enty, default_generics=None):
     return wrappers
 
 
+def typ_to_slv(typ):
+    '''
+    Convert a type to a std_logic_vector or std_logic.
+    '''
+    if isinstance(typ, typs.StdLogic):
+        new_typ = typ
+    else:
+        new_typ = typs.ConstrainedStdLogicVector(identifier=None, size=typ.size.value())
+    return new_typ
+
+
 def make_filetestbench(enty, add_double_wrapper=False, use_vunit=True,
                        default_output_path=None, default_generics=None,
-                       use_pipes=False,
+                       use_pipes=False, slv_interface=False, dut_name=None,
+                       extra_definitions='', extra_logic='',
                        ):
     '''
     Generate a testbench that reads inputs from a file, and writes outputs to
     a file.
     Args:
       `enty`: A resolved entity object parsed from the VHDL.
+      `add_double_wrapper`: Add two wrappers converting to and from std_logic_vector.  This
+         is convenient if we want to synthsize the design.
+      `use_vunit`: Make a VUnit compatiable testbench.
+      `default_output_path`: The default value for the generic that specifies the output path.
+      `default_generics`: The default values for the generics of the entity.
+      `use_pipes`: If these is True then the testbench uses named pipes for input and output
+         rather than just normal files.
+      `slv_interface`: Use if the ports in the test_module have already been converted to
+         accept std_logic and std_logic_vector.  Also assumes generics have been removed.
+      `dut_name`: The name that will be used to instantiate the dut (might be different
+                       to enty.identifier if wrappers are involved).
     '''
     if default_generics is None:
         default_generics = {}
@@ -121,11 +178,17 @@ def make_filetestbench(enty, add_double_wrapper=False, use_vunit=True,
     # Generate a record type for the entity inputs (excluding clock).
     inputs = [p for p in enty.ports.values()
               if p.direction == 'in' and p.name not in entity.CLOCK_NAMES]
-    input_names_and_types = [(p.name, p.typ) for p in inputs]
+    if slv_interface:
+        input_names_and_types = [(p.name, typ_to_slv(p.typ)) for p in inputs]
+    else:
+        input_names_and_types = [(p.name, p.typ) for p in inputs]
     input_record = typs.Record('t_input', input_names_and_types)
     # Generate a record type for the entity outputs.
     outputs = [p for p in enty.ports.values() if p.direction == 'out']
-    output_names_and_types = [(p.name, p.typ) for p in outputs]
+    if slv_interface:
+        output_names_and_types = [(p.name, typ_to_slv(p.typ)) for p in outputs]
+    else:
+        output_names_and_types = [(p.name, p.typ) for p in outputs]
     output_record = typs.Record('t_output', output_names_and_types)
     # Generate declarations and definitions for the functions to convert
     # the input and output types to and from std_logic_vector.
@@ -136,13 +199,17 @@ def make_filetestbench(enty, add_double_wrapper=False, use_vunit=True,
         package_generator.make_record_declarations_and_definitions(
             output_record))
     # Generate use clauses required by the testbench.
-    use_clauses = '\n'.join([
-        'use {}.{}.{};'.format(u.library, u.design_unit, u.name_within)
-        for u in enty.uses.values()])
-    use_clauses += '\n' + '\n'.join([
-        'use {}.{}_slvcodec.{};'.format(u.library, u.design_unit, u.name_within)
-        for u in enty.uses.values()
-        if u.library not in ('ieee', 'std') and '_slvcodec' not in u.design_unit])
+    if slv_interface:
+        use_clauses = 'use work.slvcodec.all;'
+    else:
+        use_clauses = '\n'.join([
+            'use {}.{}.{};'.format(u.library, u.design_unit, u.name_within)
+            for u in enty.uses.values()])
+        use_clauses += '\n' + '\n'.join([
+            'use {}.{}_slvcodec.{};'.format(u.library, u.design_unit, u.name_within)
+            for u in enty.uses.values()
+            if (u.library not in ('ieee', 'std')) and ('_slvcodec' not in u.design_unit) and
+            (u.design_unit not in ('slvcodec',))])
     # Get the list of generic parameters for the testbench.
     generic_params = []
     for g in enty.generics.values():
@@ -151,7 +218,10 @@ def make_filetestbench(enty, add_double_wrapper=False, use_vunit=True,
             as_str += ' := {}'.format(default_generics[g.name])
         as_str += ';'
         generic_params.append(as_str)
-    generic_params = '\n'.join(generic_params)[:-1]
+    if not slv_interface:
+        generic_params = '\n'.join(generic_params)[:-1]
+    else:
+        generic_params = ''
     # Combine the input and output record definitions with the slv conversion
     # functions.
     definitions = '\n'.join([
@@ -165,8 +235,11 @@ def make_filetestbench(enty, add_double_wrapper=False, use_vunit=True,
     connections = ',\n'.join(['{} => {}.{}'.format(
         p.name, {'in': 'input_data', 'out': 'output_data'}[p.direction], p.name)
                               for p in enty.ports.values() if p.name not in clk_names])
-    dut_generics = ',\n'.join(['{} => {}'.format(g.name, g.name)
-                               for g in enty.generics.values()])
+    if not slv_interface:
+        dut_generics = ',\n'.join(['{} => {}'.format(g.name, g.name)
+                                for g in enty.generics.values()])
+    else:
+        dut_generics = ''
     # Read in the testbench template and format it.
     if use_vunit:
         template_fn = os.path.join(os.path.dirname(__file__), 'templates',
@@ -177,10 +250,11 @@ def make_filetestbench(enty, add_double_wrapper=False, use_vunit=True,
     else:
         template_fn = os.path.join(os.path.dirname(__file__), 'templates',
                                    'file_testbench_no_vunit.vhd')
-    if add_double_wrapper:
-        dut_name = enty.identifier + '_toslvcodec'
-    else:
-        dut_name = enty.identifier
+    if dut_name is None:
+        if add_double_wrapper:
+            dut_name = enty.identifier + '_toslvcodec'
+        else:
+            dut_name = enty.identifier
     with open(template_fn, 'r') as f:
         filetestbench_template = jinja2.Template(f.read())
     filetestbench = filetestbench_template.render(
@@ -193,11 +267,16 @@ def make_filetestbench(enty, add_double_wrapper=False, use_vunit=True,
         clk_connections=clk_connections,
         connections=connections,
         output_path=default_output_path,
+        extra_definitions=extra_definitions,
+        extra_logic=extra_logic,
         )
     return filetestbench
 
 
 def process_signals(signals, type_name):
+    """
+    Generate type declarations and definitions used by the testbench.
+    """
     names_and_types = [(p.name, p.typ) for p in signals]
     record = typs.Record(type_name, names_and_types)
     slv_declarations, slv_definitions = package_generator.make_record_declarations_and_definitions(
@@ -217,7 +296,9 @@ def process_signals(signals, type_name):
 
 
 def make_use_clauses(enty):
-    # Generate use clauses required by the testbench.
+    """
+    Generate use clauses required by the testbench.
+    """
     use_clauses = '\n'.join([
         'use {}.{}.{};'.format(u.library, u.design_unit, u.name_within)
         for u in enty.uses.values()])
@@ -229,7 +310,9 @@ def make_use_clauses(enty):
 
 
 def make_generic_params(enty, default_generics=None):
-    # Get the list of generic parameters for the testbench.
+    """
+    Generate generic parameters used by the testbench.
+    """
     if default_generics is None:
         default_generics = {}
     else:
@@ -261,6 +344,12 @@ def make_filetestbench_multiple_clocks(
       `enty`: A resolved entity object parsed from the VHDL.
       `clock_domains`: An optional dictionary that maps clock_names to the
          signals in their clock domains.
+      `add_double_wrapper`: Add two wrappers converting to and from std_logic_vector.  This
+         is convenient if we want to synthsize the design.
+      `default_output_path`: The default value for the generic that specifies the output path.
+      `default_generics`: The default values for the generics of the entity.
+      `use_pipes`: If these is True then the testbench uses named pipes for input and output
+         rather than just normal files.
     '''
     grouped_ports = enty.group_ports_by_clock_domain(clock_domains)
     definitions = []
@@ -329,14 +418,19 @@ def make_filetestbench_multiple_clocks(
     return filetestbench
 
 
-def prepare_files(directory, filenames, top_entity, add_double_wrapper=False, use_vunit=True,
-                  dut_directory=None, default_generics=None, default_output_path=None,
-                  clock_domains=None, clock_periods=None, clock_offsets=None, use_pipes=False):
-    '''
+def prepare_files(
+        directory, filenames, top_entity, add_double_wrapper=False, use_vunit=True,
+        dut_directory=None, default_generics=None, default_output_path=None,
+        clock_domains=None, clock_periods=None, clock_offsets=None, use_pipes=False,
+        slv_interface=False, wrapper_name=None, extra_logic='', extra_definitions='',
+        ):
+    """
     Parses VHDL files, and generates a testbench for `top_entity`.
     Returns a tuple of a list of testbench files, and a dictionary
     of parsed objects.
-    '''
+    """
+    if wrapper_name is None:
+        wrapper_name = top_entity
     dut_fns = filenames[:]
     if dut_directory is None:
         dut_directory = directory
@@ -372,7 +466,10 @@ def prepare_files(directory, filenames, top_entity, add_double_wrapper=False, us
     else:
         ftb = make_filetestbench(resolved_entity, add_double_wrapper, use_vunit=use_vunit,
                                  default_generics=default_generics,
-                                 default_output_path=default_output_path, use_pipes=use_pipes)
+                                 default_output_path=default_output_path, use_pipes=use_pipes,
+                                 slv_interface=slv_interface, dut_name=wrapper_name,
+                                 extra_logic=extra_logic, extra_definitions=extra_definitions,
+                                 )
     ftb_fn = os.path.join(directory, '{}_tb.vhd'.format(
         resolved_entity.identifier))
     with open(ftb_fn, 'w') as f:
@@ -452,7 +549,7 @@ def add_slvcodec_files_inner(directory, filenames, packages, filename_to_package
 
 def make_add_slvcodec_files_and_setgenerics_wrapper(
         old_name, new_name, generics, ports_to_remove=None, for_arch_header='',
-        slv_interface=True):
+        slv_interface=True, wrapper_base_name='top.vhd'):
     def add_slvcodec_files_and_setgenerics_wrapper(directory, filenames):
         parsed_entities, resolved_packages, filename_to_package_name = process_files(
             directory, filenames)
@@ -461,12 +558,71 @@ def make_add_slvcodec_files_and_setgenerics_wrapper(
         parsed_entities, resolved_packages, filename_to_package_name = process_files(
             directory, combined_filenames, entity_names_to_resolve=[old_name])
         enty = parsed_entities[old_name]
-        setgenerics_wrapper = make_generics_wrapper(
+        setgenerics_wrapper, setgenerics_pkg = make_generics_wrapper(
             enty, generics, new_name, ports_to_remove, for_arch_header,
             slv_interface=slv_interface)
-        wrapper_filename = os.path.join(directory, 'top.vhd')
+        wrapper_filename = os.path.join(directory, wrapper_base_name)
+        package_filename = os.path.join(directory, 'setgenerics_pkg.vhd')
+        package_slvcodec_filename = os.path.join(directory, 'setgenerics_pkg_slvcodec.vhd')
         with open(wrapper_filename, 'w') as f:
             f.write(setgenerics_wrapper)
-        combined_filenames.append(wrapper_filename)
+        with open(package_filename, 'w') as f:
+            f.write(setgenerics_pkg)
+        with open(package_slvcodec_filename, 'w') as f:
+            f.write('''package setgenerics_pkg_slvcodec is
+            end package;''')
+        combined_filenames += [package_filename, package_slvcodec_filename, wrapper_filename]
         return combined_filenames
     return add_slvcodec_files_and_setgenerics_wrapper
+
+
+def make_formal_wrapper(enty, generics):
+    """
+    Create a wrapper around an entity which sets the generic parameters.
+    Args:
+      `enty`: A resolved entity object parsed from the VHDL.
+      `generics`: A dictionary of generics to set.
+    """
+    generics = generics.copy()
+    for k, v in generics.items():
+        if isinstance(v, str) and (len(v) > 0) and (v[0] not in  ("'", '"')):
+            generics[k] = '"' + v + '"'
+    # Get the list of generic parameters for the testbench.
+    wrapped_generics = ',\n'.join(['{} => {}'.format(g.name, g.name)
+                                   for g in enty.generics.values()])
+    # Generate use clauses required by the testbench.
+    use_clauses = '\n'.join([
+        'use {}.{}.{};'.format(u.library, u.design_unit, u.name_within)
+        for u in enty.uses.values() if (u.design_unit not in ('std_logic_1164', 'slvcodec')) and 
+                                       ('_slvcodec' not in u.design_unit)])
+    use_clauses += '\n' + '\n'.join([
+        'use {}.{}_slvcodec.{};'.format(u.library, u.design_unit, u.name_within)
+        for u in enty.uses.values()
+        if u.library not in ('ieee', 'std') and '_slvcodec' not in u.design_unit])
+
+    # Read in the pkg template and format it.
+    template_name = 'formal_pkg.vhd'
+    template_fn = os.path.join(os.path.dirname(__file__), 'templates', template_name)
+    with open(template_fn, 'r') as f:
+        template = jinja2.Template(f.read())
+    generic_infos = [{'name': g.name, 'type': g.typ, 'value': generics[g.name]}
+                     for g in enty.generics.values()]
+    pkg = template.render(
+        use_clauses=use_clauses,
+        generics=generic_infos,
+        )
+
+    use_clauses += '\nuse work.formal_pkg.all;'
+    # Read in the template and format it.
+    template_name = 'formal_wrapper.vhd'
+    template_fn = os.path.join(os.path.dirname(__file__), 'templates', template_name)
+    with open(template_fn, 'r') as f:
+        template = jinja2.Template(f.read())
+    wrapper = template.render(
+        entity_name=enty.identifier,
+        use_clauses=use_clauses,
+        wrapped_generics=wrapped_generics,
+        dut_name=enty.identifier,
+        ports=list(enty.ports.values()),
+        )
+    return pkg, wrapper
