@@ -9,10 +9,14 @@ import itertools
 import logging
 import random
 import inspect
+import json
 
+from cocotb_test import run
 from slvcodec import add_slvcodec_files
-from slvcodec import filetestbench_generator
-from slvcodec import config, fusesoc_wrapper, event
+from slvcodec import filetestbench_generator, flatten_generator
+from slvcodec import config, fusesoc_wrapper, event, cocotb_dut
+from slvcodec import cocotb_wrapper as cocotb
+from slvcodec.cocotb_wrapper import triggers
 
 
 logger = logging.getLogger(__name__)
@@ -25,10 +29,9 @@ def remove_duplicates(filenames):
             new_filenames.append(filename)
     return new_filenames
 
-
 def register_rawtest_with_vunit(
-        vu, resolved, filenames, top_entity, all_generics, test_class,
-        top_params):
+        vu, filenames, testbench_name, entity=None, resolved=None, all_generics=None,
+        test_class=None, top_params=None, pre_config=None, post_check=None):
     '''
     Register a test with vunit.
     Args:
@@ -55,18 +58,29 @@ def register_rawtest_with_vunit(
             is_unique = True
         logger.debug('Adding files to lib %s', str(filenames))
     lib.add_source_files(filenames)
-    tb_generated = lib.entity(top_entity + '_tb')
-    entity = resolved['entities'][top_entity]
+    tb_generated = lib.entity(testbench_name)
+    if all_generics is None:
+        all_generics = [{}]
     for generics_index, generics in enumerate(all_generics):
-        test = test_class(resolved, generics, top_params)
+        if (pre_config is None) or (post_check is None):
+            test = test_class(resolved, generics, top_params)
+        if pre_config is None:
+            this_pre_config = make_pre_config(test, entity, generics)
+        else:
+            this_pre_config = pre_config
+        if post_check is None:
+            this_post_check = make_post_check(test, entity, generics)
+        else:
+            this_post_check = post_check
         tb_generated.add_config(
             name=str(generics_index),
             generics=generics,
-            pre_config=make_pre_config(test, entity, generics),
-            post_check=make_post_check(test, entity, generics),
+            pre_config=this_pre_config,
+            post_check=this_post_check,
         )
 
 
+
 def register_test_with_vunit(
         vu, directory, filenames, top_entity, all_generics, test_class,
         top_params):
@@ -98,7 +112,8 @@ def register_test_with_vunit(
         vu=vu,
         resolved=resolved,
         filenames=combined_filenames,
-        top_entity=top_entity,
+        entity=resolved['entities'][top_entity],
+        testbench_name=top_entity + '_tb',
         all_generics=all_generics,
         test_class=test_class,
         top_params=top_params,
@@ -191,13 +206,14 @@ def register_coretest_with_vunit(
             generate_iteratively)
         register_rawtest_with_vunit(
             vu=vu,
-            resolved=resolved,
             filenames=combined_filenames,
-            top_entity=test['entity_name'],
+            entity=resolved['entities'][test['entity_name']],
+            testbench_name=test['entity_name'] + '_tb',
+            resolved=resolved,
             all_generics=param_set['generic_sets'],
             test_class=test['generator'],
             top_params=param_set['top_params'],
-        )
+            )
     return {
         'filenames': filenames,
         'combined_filenames': combined_filenames,
@@ -431,11 +447,12 @@ def make_old_style_test(test_generator, generics, top_params):
 
 def run_pipe_test(directory, filenames, top_entity, generics, coro, needs_resolved=False):
     simulator = event.Simulator(directory, filenames, top_entity, generics)
+    os.environ['PIPE_SIM'] = 'True'
     loop = event.EventLoop(simulator)
     if needs_resolved:
-        loop.create_task(coro(simulator.dut, loop, simulator.resolved))
+        loop.create_task(coro(simulator.dut, simulator.resolved))
     else:
-        loop.create_task(coro(simulator.dut, loop))
+        loop.create_task(coro(simulator.dut))
     loop.run_forever()
 
 
@@ -514,3 +531,140 @@ def run_coretest_with_pipes(
                 output_directory = os.path.join(test_output_directory, 'run{}'.format(dir_index))
             run_pipe_test(output_directory, filenames, test['entity_name'], generics,
                           coro, needs_resolved=needs_resolved)
+
+
+def run_coretest_with_cocotb(
+        test, test_output_directory, fusesoc_config_filename=None, generate_iteratively=False,
+        wave=False, write_input_output_files=False,
+    ):
+    '''
+    Run a test with cocotb.
+    Args:
+      `test_output_directory`: A directory in which generated files are placed.
+      `test`: A dictionary containing:
+        `param_sets`: An iteratable of top_params with lists of generics.
+        `core_name`: The name of the fusesoc core to test.
+        `wrapper_core_name`: The name of a fusesoc core that wraps the synthesizable part.
+        `top_entity`: The name of the entity to test.
+        `test_module_name`: A cocotb test module name.
+    '''
+    if 'param_sets' in test:
+        param_sets = test['param_sets']
+    else:
+        top_params = test.get('top_params', {})
+        if 'all_generics' in test:
+            all_generics = test['all_generics']
+        else:
+            all_generics = [test.get('generics', {})]
+        param_sets = [{
+            'generic_sets': all_generics,
+            'top_params': top_params,
+            }]
+    for param_set in param_sets:
+        top_params = param_set['top_params']
+
+        generated_index = 0
+        generation_directory = os.path.join(
+            test_output_directory, test['core_name'], 'generated_{}'.format(generated_index))
+        while os.path.exists(generation_directory):
+            generated_index += 1
+            generation_directory = os.path.join(
+                test_output_directory, test['core_name'], 'generated_{}'.format(generated_index))
+        os.makedirs(generation_directory)
+
+        if generate_iteratively:
+            generic_sets = param_set['generic_sets']
+            filenames = fusesoc_wrapper.generate_core_iteratively(
+                core_name=test['core_name'],
+                work_root=generation_directory,
+                all_top_generics=generic_sets,
+                top_params=top_params,
+                top_name=test['entity_name'],
+                config_filename=fusesoc_config_filename,
+                additional_generator=add_slvcodec_files,
+                )
+        else:
+            filenames = fusesoc_wrapper.generate_core(
+                working_directory=generation_directory,
+                core_name=test['core_name'],
+                parameters=top_params,
+                config_filename=fusesoc_config_filename,
+                )
+            filenames = add_slvcodec_files(generation_directory, filenames)
+        for generics in param_set['generic_sets']:
+            run_with_cocotb(
+                generation_directory, filenames, test['entity_name'],
+                generics, test['test_module_name'], test.get('test_params', None),
+                top_params, wave=wave, write_input_output_files=write_input_output_files)
+
+
+def run_with_cocotb(generation_directory, filenames, entity_name, generics, test_module_name,
+                    test_params_producer=None, top_params=None, wave=False,
+                    write_input_output_files=False, flatten=True,
+                    ):
+    entities, packages, filename_to_package_name = filetestbench_generator.process_files(
+        generation_directory, filenames, entity_names_to_resolve=entity_name)
+    combined_filenames = filetestbench_generator.add_slvcodec_files_inner(
+        generation_directory, filenames, packages, filename_to_package_name)
+    top_entity = entities[entity_name]
+    if flatten:
+        top_name = 'flat_' + entity_name
+        wrapper = flatten_generator.make_flat_wrapper(
+            top_entity, top_name, separator='_', generics=generics)
+        wrapper_filename = os.path.join(generation_directory, top_name + '.vhd')
+        with open(wrapper_filename, 'w') as f:
+            f.write(wrapper)
+        final_filenames = [wrapper_filename]
+    else:
+        top_name  = entity_name
+        final_filenames = []
+    final_filenames += combined_filenames
+    os.environ['SIM'] = 'ghdl'
+    mapping = cocotb_dut.get_entity_mapping(top_entity, generics=generics)
+    input_port_names = [port.name for port in top_entity.ports.values() if port.direction == 'in']
+    output_port_names = [port.name for port in top_entity.ports.values() if port.direction == 'out']
+    test_params_filename = os.path.abspath(os.path.join(generation_directory, 'test_params.json'))
+    if test_params_producer is not None:
+        test_params = test_params_producer({'entitites': entities, 'packages': packages})
+    else:
+        test_params = {}
+    params = {
+        'generics': generics,
+        # Use a list of tuples rather than dict so we can guarantee order.
+        'mapping': mapping,
+        'test_params': test_params,
+        'filenames': filenames,
+        'top_params': top_params,
+        }
+    if write_input_output_files:
+        params.update({
+            'input_port_names': input_port_names,
+            'output_port_names': output_port_names,
+            'directory': generation_directory,
+            })
+    with open(test_params_filename, 'w') as f:
+        f.write(json.dumps(params))
+    if wave:
+        simulation_args = ['--wave=dump.ghw']
+    else:
+        simulation_args = []
+    pwd = os.getcwd()
+    os.chdir(generation_directory)
+    run.run(
+        vhdl_sources=final_filenames,
+        simulation_args=simulation_args,
+        toplevel=top_name,
+        module=test_module_name,
+        extra_env={'test_params_filename': test_params_filename},
+        )
+    os.chdir(pwd)
+
+
+@cocotb.coroutine
+async def clock(clock_signal, period=2, units='ns'):
+    assert period % 2 == 0
+    while True:
+        clock_signal <= 0
+        await triggers.Timer(period//2, units=units)
+        clock_signal <= 1
+        await triggers.Timer(period//2, units=units)
